'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.transpile = transpile;
exports.removeDoubleEscapes = removeDoubleEscapes;
exports.compile = compile;

var _astTypes = require('ast-types');

var _recast = require('recast');

var _recast2 = _interopRequireDefault(_recast);

var _coffeeScript = require('coffee-script');

var _scope = require('coffee-script/lib/coffee-script/scope');

var _nodes = require('coffee-script/lib/coffee-script/nodes');

var _findWhere = require('lodash/collection/findWhere');

var _findWhere2 = _interopRequireDefault(_findWhere);

var _last = require('lodash/array/last');

var _last2 = _interopRequireDefault(_last);

var _flatten = require('lodash/array/flatten');

var _flatten2 = _interopRequireDefault(_flatten);

var _findIndex = require('lodash/array/findIndex');

var _findIndex2 = _interopRequireDefault(_findIndex);

var _get = require('lodash/object/get');

var _get2 = _interopRequireDefault(_get);

var _compose = require('lodash/function/compose');

var _compose2 = _interopRequireDefault(_compose);

var _isArray = require('lodash/lang/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _any = require('lodash/collection/any');

var _any2 = _interopRequireDefault(_any);

var _jscodeshift = require('jscodeshift');

var _jscodeshift2 = _interopRequireDefault(_jscodeshift);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

// regexes taken from coffeescript parser
var IS_NUMBER = /^[+-]?(?:0x[\da-f]+|\d*\.?\d+(?:e[+-]?\d+)?)$/i;
var IS_STRING = /^['"]/;
var IS_REGEX = /^\//;

function isThisMemberExpression(node) {
  return node.type === 'MemberExpression' && (node.object.type === 'ThisExpression' || node.object.name === 'this');
}

function mapBoolean(node) {
  if (node.base.val === 'true') {
    return _astTypes.builders.literal(true);
  } else if (node.base.val === 'false') {
    return _astTypes.builders.literal(false);
  }

  throwError(node.locationData, 'can\'t convert node of type: ' + node.constructor.name + ' to boolean - not recognized');
}

function stringToRegex(inputstring) {
  var match = inputstring.match(new RegExp('^/(.*?)/([gimy]*)$'));
  return new RegExp(match[1], match[2]);
}

function mapMemberProperties(properties, meta) {
  var restProperties = properties.slice(0, properties.length - 1);
  var isIndex = properties[properties.length - 1].constructor.name === 'Index';
  var right = mapExpression(properties[properties.length - 1], meta);
  var isComputed = right.type === 'Literal' || isIndex;
  var left = undefined;

  if (restProperties.length === 1) {
    left = mapExpression(restProperties[0], meta);
  } else {
    left = mapMemberProperties(restProperties, meta);
  }

  return _astTypes.builders.memberExpression(left, right, isComputed);
}

function mapMemberExpression(node, meta) {
  if ((0, _findIndex2.default)(node.base.properties, { soak: true }) > -1 || node.properties && (0, _findIndex2.default)(node.properties, { soak: true }) > -1) {
    return fallback(node, meta);
  }
  return mapMemberProperties([node.base].concat(_toConsumableArray(node.properties)), meta);
}

function mapLiteral(node) {
  var value = undefined;
  value = node.base.value;

  if (value === 'NaN') {
    return _astTypes.builders.literal(NaN);
  } else if (IS_STRING.test(value)) {
    return _astTypes.builders.literal(eval(value)); // eslint-disable-line no-eval
  } else if (IS_NUMBER.test(value)) {
      return _astTypes.builders.literal(Number(value));
    } else if (IS_REGEX.test(value)) {
      return _astTypes.builders.literal(stringToRegex(value));
    }

  return _astTypes.builders.identifier(value);
}

function mapKey(node) {
  var type = node.base.constructor.name;
  if (node.properties && node.properties.length) {
    return _astTypes.builders.identifier(node.properties[0].name.value);
  } else if (type === 'Literal') {
    return _astTypes.builders.identifier(node.base.value);
  }
}

function mapObjectExpression(node, meta) {
  return _astTypes.builders.objectExpression(node.base.properties.map(function (property) {
    return _astTypes.builders.property('init', mapExpression(property.variable || property.base, meta), mapExpression(property.value || property.base, meta));
  }));
}

function mapArrayExpression(node, meta) {
  return _astTypes.builders.arrayExpression(node.objects.map(function (expr) {
    return mapExpression(expr, meta);
  }));
}

function mapRange(node, meta) {
  var compiledRange = _recast2.default.parse(_recast2.default.prettyPrint(_recast2.default.parse(node.compile(meta)))).program.body[0];
  return compiledRange.expression;
}

function mapSlice(node, meta) {
  var jsString = node.compile(meta).substring(1);
  return _recast2.default.parse(jsString).program.body[0].expression;
}

function mapValue(node, meta) {
  var type = node.base.constructor.name;

  if (type === 'Literal') {
    return mapLiteral(node, meta);
  } else if (type === 'Range') {
    return mapRange(node, meta);
  } else if (type === 'Undefined') {
    return _astTypes.builders.identifier('undefined');
  } else if (type === 'Null') {
    return _astTypes.builders.identifier('null');
  } else if (type === 'Call') {
    return mapCall(node.base, meta);
  } else if (type === 'Bool') {
    return mapBoolean(node, meta);
  } else if (type === 'Arr' && meta.left === true) {
    return mapArrayPattern(node.base, meta);
  } else if (type === 'Obj' && meta.left === true) {
    return mapObjectPattern(node.base.properties, meta);
  } else if (type === 'Arr') {
    return mapArrayExpression(node.base, meta);
  } else if (type === 'Obj') {
    return mapObjectExpression(node, meta);
  } else if (type === 'Parens') {
    return _astTypes.builders.sequenceExpression(node.base.body.expressions.map(function (expr) {
      return mapExpression(expr, meta);
    }));
  }

  throwError(node.locationData, 'can\'t convert node of type: ' + type + ' to value - not recognized');
}

function mapOp(node, meta) {
  var operator = node.operator;

  // fall back to coffee-script modulo

  if (operator === '%%' && node.second) {
    return fallback(node, meta);
  }

  // if the cs pow operator is used, map it
  if (operator === '**') {
    return fallback(node, meta);
  }

  // fall back to coffee-script conditional operator
  if (operator === '?') {
    return fallback(node, meta);
  }

  if (operator === '++' || operator === '--') {
    return _astTypes.builders.updateExpression(operator, mapExpression(node.first, meta), !node.flip);
  }

  if (node.properties) {
    return mapExpression(node, meta);
  }

  if (node.args) {
    return mapCall(node, meta);
  }

  if (!node.second) {
    return _astTypes.builders.unaryExpression(operator, mapExpression(node.first, meta));
  }

  if (operator === '||' || operator === '&&') {
    return _astTypes.builders.logicalExpression(operator, mapExpression(node.first, meta), mapExpression(node.second, meta));
  }

  return _astTypes.builders.binaryExpression(operator, mapExpression(node.first, meta), mapExpression(node.second, meta));
}

function mapArguments(args, meta) {
  return args.map(function (arg) {
    var argName = (0, _get2.default)(arg.name, 'constructor.name');
    if (argName === 'Obj' && arg.name.objects.length === 0 || argName === 'Arr' && arg.name.objects.length === 0) {
      return _astTypes.builders.identifier(meta.scope.freeVariable('arg'));
    }

    if (arg.constructor.name === 'Expansion') {
      return _astTypes.builders.restElement(_astTypes.builders.identifier(meta.scope.freeVariable('args')));
    }

    var type = undefined;
    if (arg.name && arg.name.constructor) {
      type = arg.name.constructor.name;
    }
    if (type === 'Arr') {
      return mapArrayPattern(arg.name, meta);
    } else if (type === 'Obj') {
      return mapObjectPattern(arg.name.properties, meta);
    }

    return mapExpression(arg, meta);
  });
}

function mapCall(node, meta) {
  var left = undefined;
  var superMethodName = meta.superMethodName;

  // fallback early if variable name contains an existential operator

  if (isSoaked(node)) {
    return fallback(node, meta);
  }

  if (node.soak === true) {
    return _recast2.default.parse(node.compile(meta)).program.body[0].expression;
  } else if (node.isSuper === true && superMethodName === 'constructor') {
    left = _astTypes.builders.identifier('super');
  } else if (node.isSuper === true && superMethodName !== undefined) {
    left = _astTypes.builders.memberExpression(_astTypes.builders.identifier('super'), _astTypes.builders.identifier(superMethodName));
  } else {
    left = mapExpression(node.variable, meta);
  }

  return _astTypes.builders.callExpression(left, mapArguments(node.args, meta));
}

function mapClassProperty(node, meta) {
  return _astTypes.builders.classProperty(mapExpression(node.variable, meta), mapExpression(node.value, meta), null);
}

function mapClassBodyElement(node, meta) {
  var superMethodName = node.variable.base.value;
  var elementType = 'method';
  var isStatic = false;
  // const type = node.constructor.name;

  //  if (type === 'Assign' && node.value) {
  //    node.value.name = node.variable.base.value;
  //    node.value.variable = node.variable;
  //  }

  if (node.variable.this === true) {
    isStatic = true;
    node.variable = (0, _get2.default)(node, 'variable.properties[0].name');
  }

  if (node.constructor.name === 'Assign' && node.value && node.value.constructor.name !== 'Code') {
    if (isStatic === true) {
      return mapStaticClassProperty(node, meta);
    }
    return mapClassProperty(node, meta);
  }

  if (superMethodName === 'constructor') {
    elementType = 'constructor';
  }

  var _meta = _extends({}, meta, { isSuperMethod: true }, { superMethodName: superMethodName });

  return _astTypes.builders.methodDefinition(elementType, mapExpression(node.variable, _meta), mapExpression(node.value, _meta), isStatic);
}

function getBoundMethodNames(classElements, meta) {
  return (0, _flatten2.default)(classElements.filter(function (el) {
    return el.base && el.base.properties;
  }).map(function (el) {
    return el.base.properties;
  })).filter(function (el) {
    return (0, _get2.default)(el, 'variable.this') !== true && (0, _get2.default)(el, 'value.constructor.name') === 'Code' && el.value.bound === true;
  }).map(function (el) {
    return mapExpression(el.variable, meta);
  });
}

function unbindMethods(classElements) {
  return classElements.map(function (el) {
    if ((0, _get2.default)(el, 'value.constructor.name') === 'Code') {
      el.value.bound = false;
    }
    return el;
  });
}

function mapStaticClassProperty(node, meta) {
  var variable = (0, _get2.default)(node, 'variable.properties[0]') || node.variable;
  return _astTypes.builders.classProperty(mapExpression(variable, meta), mapExpression(node.value, meta), null, true);
}

function mapClassExpressions(expressions, meta) {
  return expressions.reduce(function (arr, expr) {
    var type = expr.constructor.name;
    var classElements = [];
    if (type === 'Assign') {
      if (expr.variable && expr.variable.this === true) {
        return arr.concat([mapStaticClassProperty(expr, meta)]);
      }
    } else if (type === 'Value') {
      classElements = expr.base.properties
      // filter out instance field variables
      .filter(function (prop) {
        return !((0, _get2.default)(prop, 'operatorToken.value') === ':' && (0, _get2.default)(prop, 'value.constructor.name') !== 'Code' && (0, _get2.default)(prop, 'variable.base.value') !== 'this');
      });
      classElements = unbindMethods(classElements);
      classElements = classElements.filter(function (el) {
        return el.constructor.name !== 'Comment';
      });
      classElements = classElements.map(function (el) {
        return mapClassBodyElement(el, meta);
      });
      return arr.concat(classElements);
    }
    return arr;
  }, []);
}

function disallowPrivateClassStatements(node) {
  if ((0, _any2.default)(node.expressions, function (expr) {
    return expr.constructor.name === 'Call' || expr.constructor.name === 'Assign' && (0, _get2.default)(expr, 'variable.this') !== true;
  })) {
    throwError(node.locationData, 'Private Class statements are not allowed.');
  }
}

function mapClassBody(node, meta) {
  var expressions = node.expressions;

  var boundMethods = getBoundMethodNames(expressions, meta);
  var classElements = mapClassExpressions(expressions, meta);
  var constructor = (0, _findWhere2.default)(classElements, { kind: 'constructor' });

  disallowPrivateClassStatements(node);

  if (boundMethods.length > 0) {
    if (constructor === undefined) {
      // create an empty constructor if there isn't one yet
      constructor = _astTypes.builders.methodDefinition('constructor', _astTypes.builders.identifier('constructor'), _astTypes.builders.functionExpression(null, [], _astTypes.builders.blockStatement([])));
      classElements.unshift(constructor);
    }

    // bind all the bound methods to the class
    var body = constructor.value.body.body;
    var hasSuper = !!(0, _findWhere2.default)(body, {
      expression: {
        callee: {
          name: 'super'
        }
      }
    });

    body.splice.apply(body, [hasSuper ? 1 : 0, 0].concat(_toConsumableArray(boundMethods.map(function (identifier) {
      return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', _astTypes.builders.memberExpression(_astTypes.builders.thisExpression(), identifier), _astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.memberExpression(_astTypes.builders.thisExpression(), identifier), _astTypes.builders.identifier('bind')), [_astTypes.builders.thisExpression()])));
    }))));
  }

  return _astTypes.builders.classBody(classElements);
}

function mapClassExpression(node, meta) {
  // if this is an anonymous class expression fallback
  // to the cs compiler
  if (node.variable === undefined && node.parent === undefined && node.body.expressions.length < 1) {
    return fallback(node, meta);
  }

  var parent = null;

  if (node.parent !== undefined && node.parent !== null) {
    parent = mapExpression(node.parent, meta);
  }

  return _astTypes.builders.classExpression(mapExpression(node.variable, meta), mapClassBody(node.body, meta), parent);
}

function mapClassDeclaration(node, meta) {
  var parent = null;

  if (node.variable) {
    node.ensureConstructor(node.variable.base.value);
  }

  var code = new _nodes.Code([], _nodes.Block.wrap([node.body]));
  meta = _extends({}, meta, { classScope: code.makeScope(meta.scope) });

  if ((0, _get2.default)(node, 'variable.properties.length') > 0) {
    return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', mapExpression(node.variable, meta), mapClassExpression(_extends({}, node, { variable: (0, _last2.default)(node.variable.properties) }), meta)));
  }

  if (node.parent !== undefined && node.parent !== null) {
    parent = mapExpression(node.parent, meta);
    meta = _extends({}, meta, { extendedClass: true });
  }

  if (!node.variable) {
    return _astTypes.builders.expressionStatement(_astTypes.builders.parenthesizedExpression(_astTypes.builders.classExpression(null, mapClassBody(node.body, meta), parent)));
  }

  return _astTypes.builders.classDeclaration(mapExpression(node.variable, meta), mapClassBody(node.body, meta), parent);
}

function mapElseBlock(node, meta) {
  var type = node.constructor.name;

  if (type === 'If') {
    var conditional = mapIfStatement(node, meta);
    if (_astTypes.namedTypes.IfStatement.check(conditional)) {
      return conditional;
    }
    return _astTypes.builders.blockStatement([conditional]);
  } else if (type === 'Block') {
    return mapBlockStatement(node, meta);
  }

  return mapBlockStatement({ expressions: [node] }, meta);
}

function mapElseExpression(node, meta) {
  var type = node.constructor.name;

  if (type === 'If') {
    return mapConditionalExpression(node, meta);
  } else if (type === 'Block') {
    return mapExpression(node.expressions[0], meta);
  }

  return mapExpression(node, meta);
}

function mapConditionalExpression(node, meta) {
  var alternate = _astTypes.builders.identifier('undefined');

  if (node.elseBody) {
    alternate = mapElseExpression(node.elseBody.expressions[0], meta);
  }

  return _astTypes.builders.conditionalExpression(mapExpression(node.condition, meta), mapExpression(node.body.expressions[0], meta), alternate);
}

function mapTryExpression(node, meta) {
  var tryBlock = mapTryCatchBlock(node, meta);
  tryBlock.block = addReturnStatementToBlock(tryBlock.block);
  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([tryBlock])), []);
}

function mapIfStatement(node, meta) {
  var alternate = null;
  var elseBody = node.elseBody;

  // The coffeescript doesn't explicitly tell you if something is
  // an if-else block so we need to make some checks and then a little
  // plumbing to put this in the right place.
  if ((0, _get2.default)(elseBody, 'expressions.length') === 1 && (0, _get2.default)(elseBody, 'expressions[0].constructor.name') === 'If') {
    elseBody = elseBody.expressions[0];
  }

  if (elseBody) {
    alternate = mapElseBlock(elseBody, meta);
  }

  return _astTypes.builders.ifStatement(mapExpression(node.condition, meta), mapBlockStatement(node.body, meta), alternate);
}

function isTernaryOperation(node) {
  var regex = /^(Literal|Code)/;

  return (0, _get2.default)(node, 'body.expressions.length') === 1 && regex.test((0, _get2.default)(node, 'body.expressions[0].base.constructor.name')) && regex.test((0, _get2.default)(node, 'elseBody.expressions[0].base.constructor.name')) && (0, _get2.default)(node, 'elseBody.expressions.length') === 1;
}

function mapConditionalStatement(node, meta) {
  if (isTernaryOperation(node)) {
    return _astTypes.builders.expressionStatement(mapConditionalExpression(node, meta));
  }
  return mapIfStatement(node, meta);
}

function mapTryCatchBlock(node, meta) {
  var finalize = null;
  var catchBlock = null;
  if (node.recovery) {
    var recovery = mapBlockStatement(node.recovery, meta);
    var errorVar = mapLiteral({ base: node.errorVariable }, meta);

    catchBlock = _astTypes.builders.catchClause(errorVar, null, recovery);
  }

  if (node.ensure) {
    finalize = mapBlockStatement(node.ensure, meta);
  } else if (!catchBlock) {
    finalize = _astTypes.builders.blockStatement([]);
  }

  return _astTypes.builders.tryStatement(mapBlockStatement(node.attempt, meta), catchBlock, finalize);
}

function mapReturnStatement(node, meta) {
  return _astTypes.builders.returnStatement(node.expression ? mapExpression(node.expression, meta) : null);
}

function isSoaked(node) {
  return node.variable && (0, _findIndex2.default)((0, _get2.default)(node, 'variable.base.properties'), { soak: true }) > -1 || node.variable && node.variable.properties && (0, _findIndex2.default)(node.variable.properties, { soak: true }) > -1;
}

function mapStatement(node, meta) {
  var type = node.constructor.name;

  if (type === 'While') {
    return mapWhileLoop(node, meta);
  } else if (type === 'Return') {
    return mapReturnStatement(node, meta);
  } else if (type === 'Throw') {
    return mapThrowStatement(node, meta);
  } else if (type === 'Comment') {
    return _astTypes.builders.emptyStatement();
  } else if (type === 'For') {
    return mapForStatement(node, meta);
  } else if (type === 'Class') {
    return mapClassDeclaration(node, meta);
  } else if (type === 'Switch') {
    return mapSwitchStatement(node, meta);
  } else if (type === 'If') {
    return mapConditionalStatement(node, meta);
  } else if (type === 'Try') {
    return mapTryCatchBlock(node, meta);
  }

  return _astTypes.builders.expressionStatement(mapExpression(node, meta));
}

function mapBlockStatements(node, meta) {
  return (0, _flatten2.default)(node.expressions.map(function (expr) {
    var type = expr.constructor.name;
    var prototypeProps = [];
    if (type === 'Class') {
      // extract prototype assignments
      prototypeProps = (0, _flatten2.default)(expr.body.expressions.filter(function (ex) {
        return ex.constructor.name === 'Value';
      }).map(function (ex) {
        return ex.base.properties;
      })).filter(function (ex) {
        return (0, _get2.default)(ex, 'operatorToken.value') === ':' && (0, _get2.default)(ex, 'value.constructor.name') !== 'Code' && (0, _get2.default)(ex, 'variable.base.value') !== 'this';
      }).filter(function (ex) {
        return (0, _get2.default)(ex, 'value.constructor.name') !== 'Code';
      }).map(function (ex) {
        return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', _astTypes.builders.memberExpression(_astTypes.builders.memberExpression(mapExpression(expr.variable), _astTypes.builders.identifier('prototype')), mapExpression(ex.variable, meta)), mapExpression(ex.value, meta)));
      });
    }

    return [mapStatement(expr, meta)].concat(prototypeProps);
  }));
}

function addVariablesToScope() {
  var nodes = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
  var meta = arguments[1];
  var context = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];

  // recursively  add all variables to the cs scope object to
  // prevent any naming collisions that might occur when the
  // coffee-script compiler needs to generate variable names
  nodes.forEach(function (node) {
    var type = node.constructor.name;

    if (type === 'Param') {
      var nameType = node.name.constructor.name;

      if (nameType === 'Obj' || nameType === 'Arr') {
        addVariablesToScope(node.name.objects, meta, true);
      } else if (nameType === 'Literal') {
        meta.scope.add(node.name.value, 'var');
      }
    } else if (type === 'Code') {
      addVariablesToScope(node.params, meta, true);
    } else if (type === 'Assign') {
      var varType = node.variable.base.constructor.name;

      if (varType === 'Literal') {
        meta.scope.add(node.variable.base.value, 'var');
      }

      if (varType === 'Obj' || varType === 'Arr') {
        addVariablesToScope(node.variable.base.objects, meta, true);
      }

      if (node.context === 'object' && node.value && node.value.base.objects) {
        addVariablesToScope(node.value.base.objects, meta, true);
      }
    } else if (type === 'Value' && context === true) {
      meta.scope.add(node.base.value, 'var');
    }
  });
}

function mapBlockStatement(node, meta) {
  var factory = arguments.length <= 2 || arguments[2] === undefined ? _astTypes.builders.blockStatement : arguments[2];

  addVariablesToScope(node.expressions, meta);
  var block = factory(mapBlockStatements(node, meta));
  return block;
}

function mapInArrayExpression(node, meta) {
  var test = _astTypes.builders.memberExpression(mapExpression(node.array, meta), _astTypes.builders.callExpression(_astTypes.builders.identifier('includes'), [mapExpression(node.object, meta)]));

  if (node.negated) {
    test = _astTypes.builders.unaryExpression('!', test);
  }

  return test;
}

function extractAssignStatementsByArguments(nodes) {
  function mapPatternThisAssignmentsToMemberExpressions(node) {
    return node.properties.filter(function (assignment) {
      return assignment.value.name === 'this';
    }).map(function (assignment) {
      return _astTypes.builders.memberExpression(_astTypes.builders.thisExpression(), _astTypes.builders.identifier(assignment.key.name));
    });
  }

  return (0, _flatten2.default)(nodes.map(function (node) {
    return node.type === 'AssignmentExpression' ? node.left : node;
  }).map(function (node) {
    return node.type === 'RestElement' ? node.argument : node;
  }).map(function (node) {
    return node.type === 'ObjectPattern' ? mapPatternThisAssignmentsToMemberExpressions(node) : node;
  })).filter(isThisMemberExpression).map(function (node) {
    return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', node, node.property));
  });
}

function normalizeArgument(node) {
  if (node.type === 'AssignmentExpression' && node.left.type === 'MemberExpression') {
    return _astTypes.builders.assignmentExpression(node.operator, node.left.property, node.right);
  }
  if (isThisMemberExpression(node)) {
    return { type: 'Identifier', name: node.property.name };
  }
  return node;
}

function normalizeArguments(nodes) {
  return nodes.map(function (node) {
    if (node.type === 'RestElement') {
      node.argument = normalizeArgument(node.argument);
      return node;
    }
    return normalizeArgument(node);
  });
}

function transformToExpression(_node) {
  var node = _node;

  if (node.expression !== undefined) {
    return node.expression;
  }

  if (node.type === 'IfStatement') {
    node = addReturnStatementToIfBlocks(node);
  } else if (node.tpye === 'SwitchStatement') {
    node = node;
  }

  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([node])), []);
}

function lastReturnStatement() {
  var nodeList = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  if (nodeList.length > 0) {
    var lastIndex = nodeList.length - 1;

    if (nodeList[lastIndex].type === 'SwitchStatement') {
      nodeList[lastIndex] = addReturnStatementsToSwitch(nodeList[lastIndex]);
    } else if (nodeList[lastIndex].type === 'ThrowStatement') {
      return nodeList;
    } else if (nodeList[lastIndex].type === 'IfStatement') {
      nodeList[lastIndex] = addReturnStatementToIfBlocks(nodeList[lastIndex]);
    } else if (nodeList[lastIndex].type === 'TryStatement') {
      nodeList[lastIndex] = addReturnStatementsToTryCatch(nodeList[lastIndex]);
    } else {
      nodeList[lastIndex] = _astTypes.builders.returnStatement(transformToExpression(nodeList[nodeList.length - 1]));
    }
  }
  return nodeList;
}

function lastBreakStatement() {
  var nodeList = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  var returns = nodeList.filter(function (node) {
    return node.type === 'ReturnStatement';
  });
  if (returns.length < 1 && nodeList.length > 0) {
    nodeList.push(_astTypes.builders.breakStatement());
  }
  return nodeList;
}

function addReturnStatementToIfBlocks(node) {
  node.consequent = addReturnStatementToBlock(node.consequent);
  if (_astTypes.namedTypes.IfStatement.check(node.alternate)) {
    node.alternate = addReturnStatementToIfBlocks(node.alternate);
  } else if (_astTypes.namedTypes.ExpressionStatement.check(node.alternate)) {
    node.alternate = _astTypes.builders.returnStatement(node.alternate.expression);
  } else if (node.alternate) {
    node.alternate = addReturnStatementToBlock(node.alternate);
  }
  return node;
}

function addReturnStatementToBlock(node) {
  var hasReturnStatement = (0, _findIndex2.default)(node.body, { type: 'ReturnStatement' }) === node.body.length - 1;
  if (!hasReturnStatement) {
    node.body = lastReturnStatement(node.body);
  }
  return node;
}

function addReturnStatementsToTryCatch(node) {
  node.block = addReturnStatementToBlock(node.block);
  if (node.handler) {
    node.handler.body = addReturnStatementToBlock(node.handler.body);
  }

  return node;
}

function detectIllegalSuper(node, meta) {
  var superIndex = (0, _findIndex2.default)((0, _get2.default)(node, 'body.expressions'), { isSuper: true });
  var hasArgumentAssignments = (0, _any2.default)(node.params, { name: { this: true } });
  var isConstructor = meta.superMethodName === 'constructor';
  var isExtendedClass = meta.extendedClass;
  var firstThisAssignmentIndex = (0, _findIndex2.default)((0, _get2.default)(node, 'body.expressions'), { variable: { this: true } });
  var superCall = (0, _get2.default)(node, 'body.expressions')[superIndex];

  var hasArgumentAssignmentsAndSuperCall = isExtendedClass && isConstructor && hasArgumentAssignments && superIndex > -1;

  var hasSuperCallAfterThisAssignments = isExtendedClass && isConstructor && firstThisAssignmentIndex > -1 && superIndex > firstThisAssignmentIndex;

  if (hasArgumentAssignmentsAndSuperCall || hasSuperCallAfterThisAssignments) {
    throwError(superCall.locationData, 'Illegal use of super() in constructor. super must be called before any this assignments');
  }
}

function throwError(locData, msg) {
  throw new Error('[' + locData.first_line + ':' + locData.first_column + '] - ' + msg);
}

function mapFunction(node, meta) {
  // Function {
  //   params: [],
  //   body: [statements],
  //   bound: Boolean
  // }
  var isGenerator = node.isGenerator;
  var isConstructor = meta.superMethodName === 'constructor' && meta.isSuperMethod;

  // throw an error when there's an illegal super statement
  detectIllegalSuper(node, meta);

  meta = _extends({}, meta, { scope: node.makeScope(meta.scope) }, { isSuperMethod: false });

  var args = mapArguments(node.params, meta);
  // restIndex is the location of the splat argument
  var restIndex = (0, _findIndex2.default)(args, _astTypes.namedTypes.RestElement.check);
  // In coffeescript you can immediately assign an argument to a
  // member of `this`. Which looks like this: fn = (@a = 'A') ->
  // For our compilation we translate it like
  // fn = function(a) { this.a = a; } as there is no 1 to 1
  // solution here
  // setupStatements will be appended at the top of the function
  // block. It's used to add behaviour that would be impossible to
  // map 1 to 1 from coffeescript
  var setupStatements = extractAssignStatementsByArguments(args, meta);
  // Remove any assignments to this, as those are in setupStatements by now
  args = normalizeArguments(args, meta);

  // In CoffeeScript you can have arguments after the rest argument, as a 'tail' of sorts
  // This is not possible in es2015, so instead we change something like: fn = (a, b..., c) ->
  // To: fn = function(a, ...b) { var [c] = b.splice(Math.max(0, b.length - 1)); }
  if (restIndex !== -1 && restIndex < args.length - 1) {
    var tailArgs = args.splice(restIndex + 1, args.length - restIndex - 1);
    var name = args[restIndex].argument.name;
    var tailStatements = [];
    tailArgs.forEach(function (arg) {
      if (arg.type === 'AssignmentExpression') {
        arg.type = 'AssignmentPattern';
      }
    });
    tailStatements.unshift(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(_astTypes.builders.arrayPattern(tailArgs), _astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.identifier(name), _astTypes.builders.identifier('splice')), [_astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.identifier('Math'), _astTypes.builders.identifier('max')), [_astTypes.builders.literal(0), _astTypes.builders.binaryExpression('-', _astTypes.builders.memberExpression(_astTypes.builders.identifier(name), _astTypes.builders.identifier('length')), _astTypes.builders.literal(tailArgs.length))])]))]));
    setupStatements = tailStatements.concat(setupStatements);
  }

  // since we are going to be using an arrow function, we can throw away the special
  // context that CoffeeScript created for us
  meta.scope.method.context = 'this';

  var block = mapBlockStatement(node.body, meta);

  if (isGenerator === false && !isConstructor) {
    block = addReturnStatementToBlock(block, meta);
  }

  block.body = setupStatements.concat(block.body);

  if (node.bound === true) {
    return _astTypes.builders.arrowFunctionExpression(args, block);
  }

  return _astTypes.builders.functionExpression(null, args, block, isGenerator);
}

function insertSuperCall(path) {
  var classMethods = (0, _get2.default)(path, 'value.body.body') || [];
  var constructorIndex = (0, _findIndex2.default)(classMethods, { kind: 'constructor' });
  if (constructorIndex > -1) {
    var superCalls = (0, _jscodeshift2.default)(classMethods[constructorIndex]).find(_jscodeshift2.default.CallExpression, { callee: { name: 'super' } }).nodes();
    if (superCalls.length < 1) {
      classMethods[constructorIndex].value.body.body.unshift(_astTypes.builders.expressionStatement(_astTypes.builders.callExpression(_astTypes.builders.identifier('super'), [_astTypes.builders.spreadElement(_astTypes.builders.identifier('arguments'))])));
    }
  }
}

function insertSuperCalls(ast) {
  (0, _jscodeshift2.default)(ast).find(_jscodeshift2.default.ClassDeclaration, function (path) {
    return (0, _get2.default)(path, 'superClass');
  }).forEach(insertSuperCall);

  (0, _jscodeshift2.default)(ast).find(_jscodeshift2.default.ClassExpression, function (path) {
    return (0, _get2.default)(path, 'superClass');
  }).forEach(insertSuperCall);

  return ast;
}

function getDestructuringAssignmentInfo(path) {
  var assignmentPath = path.value.left;
  var targetIds = (0, _jscodeshift2.default)(assignmentPath).find(_jscodeshift2.default.Identifier).filter(function (_path) {
    return _path.parentPath.name === 'elements';
  }).paths().map(function (idPath) {
    return idPath.value.name;
  });

  return {
    targetIds: targetIds,
    undeclaredIds: targetIds.filter(function (id) {
      return !path.scope.lookup(id);
    }),
    hasMemberAssignments: (0, _jscodeshift2.default)(assignmentPath).find('MemberExpression').size() > 0
  };
}

function findNodeParent(node, matcher) {
  if (node.parent) {
    if (matcher(node.parent)) {
      return node.parent;
    }
    return findNodeParent(node.parent, matcher);
  }
}

function insertBreakStatements(ast) {
  (0, _jscodeshift2.default)(ast)
  // find all switch statements
  .find(_jscodeshift2.default.SwitchStatement).forEach(function (node) {
    return (0, _jscodeshift2.default)(node).replaceWith(function (path) {
      return addBreakStatementsToSwitch(path.value);
    });
  });
  return ast;
}

function insertDestructuringAssignmentVars(path) {
  var _getDestructuringAssi = getDestructuringAssignmentInfo(path);

  var hasMemberAssignments = _getDestructuringAssi.hasMemberAssignments;
  var targetIds = _getDestructuringAssi.targetIds;
  var undeclaredIds = _getDestructuringAssi.undeclaredIds;
  var _path$value = path.value;
  var left = _path$value.left;
  var right = _path$value.right;


  if (hasMemberAssignments || undeclaredIds.length !== targetIds.length) {
    undeclaredIds.forEach(function (id) {
      return path.scope.injectTemporary(_jscodeshift2.default.identifier(id), null);
    });
  } else {
    (0, _jscodeshift2.default)(path).replaceWith(function () {
      return _astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(_jscodeshift2.default.arrayPattern(left.elements), right)]);
    });
  }
  path.scope.scan(true);
}

function insertVariableDeclarations(ast) {
  (0, _jscodeshift2.default)(ast)
  // first we're going to find all assignments in our code
  .find(_jscodeshift2.default.AssignmentExpression, function (node) {
    return !_astTypes.namedTypes.MemberExpression.check(node.left) && (0, _get2.default)(node, 'operator') === '=';
  })
  // then we're going to search for the uppermost assignment
  // of similarly named variables
  .forEach(function (path) {
    // There's something weird with AssignmentExpressions vs. AssignmentPatterns in function arguments that
    // causes scope.lookup to not find the definition properly.
    if (path.parent.value.type.match(/^(ArrowFunctionExpression|Function(Expression|Statement))$/)) {
      return;
    }

    if (_jscodeshift2.default.ArrayExpression.check(path.value.left)) {
      insertDestructuringAssignmentVars(path);
      return;
    }

    var needle = path.value.left.name;

    if (!path.scope.lookup(needle)) {
      var blockNode = findNodeParent(path, function (node) {
        return (0, _get2.default)(node, 'value.type') === 'BlockStatement';
      });
      if (path.parent && (0, _get2.default)(path, 'parent.parent.value.type') !== 'SwitchCase' && path.parent.value.type === 'ExpressionStatement' && (0, _get2.default)(blockNode, 'parent.value.type') !== 'IfStatement') {
        (0, _jscodeshift2.default)(path).replaceWith(function (_path) {
          return _astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(_path.value.left, _path.value.right)]);
        });
      } else {
        path.scope.injectTemporary(path.value.left);
      }
      path.scope.scan(true);
    }
  });
  return ast;
}

function mapSwitchCases(cases, meta) {
  return cases.reduce(function (arr, nodes) {
    var tests = (0, _isArray2.default)(nodes[0]) ? nodes[0] : [nodes[0]];
    var switchCases = tests.filter(function (expr) {
      return !!expr;
    }).map(function (expr) {
      return _astTypes.builders.switchCase(mapExpression(expr), []);
    });
    var consequent = mapBlockStatements((0, _get2.default)(nodes, 1), meta);

    if (switchCases.length > 0) {
      switchCases[switchCases.length - 1].consequent = consequent;
    } else if (switchCases.length === 0 && !!nodes[1]) {
      switchCases.push(_astTypes.builders.switchCase(null, consequent));
    }

    return arr.concat((0, _flatten2.default)(switchCases));
  }, []);
}

function mapSwitchStatement(node, meta) {
  var cases = [];
  var subject = undefined;

  if (node.cases && node.cases.length > 0) {
    cases = cases.concat(node.cases);
  }

  if (node.otherwise) {
    cases.push([null, node.otherwise]);
  }

  cases = mapSwitchCases(cases, meta);

  if (node.subject) {
    subject = mapExpression(node.subject, meta);
  } else {
    cases = cases.map(function (cas) {
      if (cas.test !== null) {
        cas.test = _astTypes.builders.unaryExpression('!', cas.test);
      }
      return cas;
    });
    subject = _astTypes.builders.literal(false);
  }

  return _astTypes.builders.switchStatement(subject, cases);
}

function mapForGuard(guardNode, meta) {
  var isExistential = guardNode.constructor.name === 'Existence';
  var guardClause = isExistential ? mapExistentialExpression(guardNode, meta) : mapOp(guardNode.expression || guardNode, meta);

  return guardClause;
}

function mapForStatement(node, meta) {
  var blockStatement = mapBlockStatement(node.body, meta);

  // wrap blockStatement in a conditional if there
  // is a conditional expression attached to the for
  // loop
  if (node.guard) {
    blockStatement = _astTypes.builders.blockStatement([_astTypes.builders.ifStatement(mapForGuard(node.guard, meta), blockStatement)]);
  }

  if (node.object === false) {
    if (node.index === undefined) {
      var name = node.name === undefined ? _astTypes.builders.identifier('_i') : mapExpression(node.name, _extends({}, meta, { left: true }));
      return _astTypes.builders.forOfStatement(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(name, null)]), mapExpression(node.source, meta), blockStatement);
    }
    return _astTypes.builders.forOfStatement(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(_astTypes.builders.arrayPattern([mapExpression(node.index, meta), mapExpression(node.name, meta)]), null)]), _astTypes.builders.callExpression(_astTypes.builders.memberExpression(mapExpression(node.source, meta), _astTypes.builders.identifier('entries')), []), blockStatement);
  } else if (node.object === true) {
    var declaration = undefined;
    var method = undefined;
    if (node.name === undefined) {
      declaration = mapExpression(node.index, meta);
      method = 'keys';
    } else {
      declaration = _astTypes.builders.arrayPattern([mapExpression(node.index, meta), mapExpression(node.name, meta)]);
      method = 'entries';
    }

    return _astTypes.builders.forOfStatement(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(declaration, null)]), _astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.identifier('Object'), _astTypes.builders.identifier(method)), [mapExpression(node.source, meta)]), blockStatement);
  }
}

function mapLeftHandForExpression(node, meta) {
  if (node.step !== undefined) {
    return _astTypes.builders.memberExpression(mapExpression(node.source, meta), _astTypes.builders.callExpression(_astTypes.builders.identifier('filter'), [_astTypes.builders.arrowFunctionExpression([_astTypes.builders.identifier('_'), _astTypes.builders.identifier('_i')], _astTypes.builders.blockStatement([_astTypes.builders.returnStatement(_astTypes.builders.logicalExpression('||', _astTypes.builders.binaryExpression('===', _astTypes.builders.identifier('_i'), _astTypes.builders.literal(0)), _astTypes.builders.binaryExpression('===', _astTypes.builders.binaryExpression('%', _astTypes.builders.identifier('_i'), _astTypes.builders.binaryExpression('+', mapExpression(node.step, meta), _astTypes.builders.literal(1))), _astTypes.builders.literal(0))))], _recast2.default.parse('return _i === 0 || _i % (2 + 1) == 0;').program.body))]));
  }

  return mapExpression(node.source, meta);
}

function mapForGuardToFilter(guardNode, target, args, meta) {
  if (!guardNode) {
    return null;
  }

  var guardClause = mapForGuard(guardNode, meta);
  return _astTypes.builders.callExpression(_astTypes.builders.memberExpression(target, _astTypes.builders.identifier('filter')), [_astTypes.builders.arrowFunctionExpression(args, guardClause)]);
}

function mapForExpression(node, meta) {
  var leftHand = mapLeftHandForExpression(node, meta);
  var args = [];
  var target = undefined;
  if (node.object === true) {
    var method = undefined;
    if (node.name === undefined) {
      args.push(mapExpression(node.index, meta));
      method = 'keys';
    } else {
      args.push(_astTypes.builders.arrayPattern([mapExpression(node.index, meta), mapExpression(node.name, meta)]));
      method = 'entries';
    }
    target = _astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.identifier('Object'), _astTypes.builders.identifier(method)), [leftHand]);
  } else {
    target = leftHand;
    if (node.name !== undefined) {
      args.push(mapExpression(node.name, meta));
    }
    if (node.index !== undefined) {
      args.push(mapExpression(node.index, meta));
    }
  }

  var blockStatement = mapBlockStatement(node.body, meta);
  var filterCall = mapForGuardToFilter(node.guard, target, args, meta);
  if (filterCall) {
    var hasNoLoopTransformLogic = blockStatement.body.length === 1 && _astTypes.namedTypes.Identifier.check(blockStatement.body[0].expression);

    if (hasNoLoopTransformLogic) {
      return filterCall;
    }
  }

  return _astTypes.builders.callExpression(_astTypes.builders.memberExpression(filterCall || target, _astTypes.builders.identifier('map')), [_astTypes.builders.arrowFunctionExpression(args, addReturnStatementToBlock(blockStatement))]);
}

function mapSplatParam(node, meta) {
  return _astTypes.builders.restElement(mapExpression(node, meta));
}

function mapParam(node, meta) {
  if (node.value !== undefined && node.value !== null) {
    return mapExpression(mapParamToAssignment(node), meta);
  } else if (node.splat === true) {
    return mapSplatParam(node.name, meta);
  }
  return mapExpression(node.name, meta);
}

function mapSplat(node, meta) {
  return _astTypes.builders.spreadElement(mapExpression(node.name, meta));
}

function addReturnStatementsToSwitch(node) {
  node.cases = node.cases.map(function (switchCase) {
    switchCase.consequent = lastReturnStatement(switchCase.consequent);
    return switchCase;
  });
  return node;
}

function addBreakStatementsToSwitch(node) {
  node.cases = node.cases.map(function (switchCase, index) {
    var isLastCase = index + 1 === node.cases.length;
    if (switchCase.test !== null && !isLastCase) {
      switchCase.consequent = lastBreakStatement(switchCase.consequent);
    }
    return switchCase;
  });
  return node;
}

function mapSwitchExpression(node, meta) {
  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([addReturnStatementsToSwitch(mapSwitchStatement(node, meta))])), []);
}

function fallback(node, meta) {
  var compiled = node.compile(meta);
  return _recast2.default.parse(_recast2.default.prettyPrint(_recast2.default.parse(compiled), meta.options)).program.body[0].expression;
}

function mapExistentialExpression(node, meta) {
  return fallback(node, meta);
}

function mapNewExpression(node, meta) {
  var constructor = node.first || node.variable;
  var args = node.args ? mapArguments(node.args, meta) : [];
  return _astTypes.builders.newExpression(mapExpression(constructor, meta), args);
}

function conditionalStatementAsExpression(node, meta) {
  var conditionalStatement = mapConditionalStatement(node, meta);

  if (conditionalStatement.type === 'IfStatement') {
    return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([addReturnStatementToIfBlocks(conditionalStatement)])), []);
  }

  return conditionalStatement.expression;
}

// function mapComment(node) {
//   const comment = b.block(node.comment);
//   return comment;
// }

function mapWhileLoop(node, meta) {
  return _astTypes.builders.whileStatement(mapExpression(node.condition, meta), mapBlockStatement(node.body, meta));
}

function mapThrowStatement(node, meta) {
  return _astTypes.builders.throwStatement(mapExpression(node.expression, meta));
}

function mapWhileExpression(node, meta) {
  return fallback(node, meta);
}

function mapYieldExpression(node, meta) {
  return _astTypes.builders.yieldExpression(mapExpression(node.first, meta));
}

function mapExpression(node, meta) {
  var type = node.constructor.name;

  if (node.properties && node.properties.length > 0) {
    return mapMemberExpression(node, meta);
  } else if (type === 'Range') {
    return mapRange(node, meta);
  } else if (type === 'If') {
    return conditionalStatementAsExpression(node, meta);
  } else if (type === 'Parens' && (0, _get2.default)(node, 'body.expressions[0]')) {
    return _astTypes.builders.parenthesizedExpression(mapExpression((0, _get2.default)(node, 'body.expressions[0]'), meta));
  } else if (type === 'Parens') {
    return _astTypes.builders.parenthesizedExpression(mapExpression(node.body, meta));
  } else if (type === 'Arr') {
    return mapValue({ base: node }, meta);
  } else if (type === 'Try') {
    return mapTryExpression(node, meta);
  } else if (type === 'Call' && node.isNew === true) {
    return mapNewExpression(node, meta);
  } else if (type === 'Op' && node.operator === 'yield') {
    return mapYieldExpression(node, meta);
  } else if (type === 'Op' && node.operator === 'new') {
    return mapNewExpression(node, meta);
  } else if (type === 'Existence') {
    return mapExistentialExpression(node, meta);
  } else if (type === 'While') {
    return mapWhileExpression(node, meta);
  } else if (type === 'Switch') {
    return mapSwitchExpression(node, meta);
  } else if (type === 'Splat') {
    return mapSplat(node, meta);
  } else if (type === 'Assign') {
    return mapAssignmentExpression(node, meta);
  } else if (type === 'Slice') {
    return mapSlice(node, meta);
  } else if (type === 'For') {
    return mapForExpression(node, meta);
  } else if (type === 'Param') {
    return mapParam(node, meta);
  } else if (type === 'Class') {
    return mapClassExpression(node, meta);
  } else if (type === 'Extends' && node.parent && node.child) {
    return fallback(node, meta);
  } else if (type === 'Extends') {
    return mapExpression(node.parent, meta);
  } else if (type === 'Code') {
    // Code is just a stupid word for function
    return mapFunction(node, meta);
  } else if (type === 'Index') {
    return mapExpression(node.index, meta);
  } else if (type === 'Access') {
    return mapExpression(node.name, meta);
  } else if (type === 'Literal') {
    return mapLiteral({ base: node }, meta);
  } else if (type === 'In') {
    return mapInArrayExpression(node, meta);
  } else if (type === 'Value') {
    return mapValue(node, meta);
  } else if (type === 'Op') {
    return mapOp(node, meta);
  } else if (type === 'Call') {
    return mapCall(node, meta);
  }

  throwError(node.locationData, 'can\'t convert node of type: ' + type + ' to Expression - not recognized');
}

function mapParamToAssignment(node) {
  var assignment = {
    variable: node.name,
    value: node.value
  };
  assignment.constructor = { name: 'Assign' };
  return assignment;
}

function mapExistentialAssignmentExpression(node, meta) {
  var unsoaked = node.variable.unfoldSoak(meta);
  var condition = mapExistentialExpression(unsoaked.condition, meta);
  var assignTarget = mapExpression(unsoaked.body, meta);
  var assignValue = mapExpression(node.value, meta);
  var operator = node.context || '=';

  return _astTypes.builders.conditionalExpression(condition, _astTypes.builders.assignmentExpression(operator, assignTarget, assignValue), _astTypes.builders.unaryExpression('void', _astTypes.builders.literal(0)));
}

function mapAssignmentExpression(node, meta) {
  if (isSoaked(node)) {
    return mapExistentialAssignmentExpression(node, meta);
  }

  var variable = undefined;
  var props = (0, _get2.default)(node, 'variable.base.properties') || [];
  if ((0, _any2.default)(props, { this: true })) {
    return fallback(node, meta);
  }
  if ((0, _get2.default)(node, 'variable.base.properties.length') > 0) {
    variable = mapAssignmentPattern(node.variable.base, meta);
  } else {
    variable = mapExpression(node.variable, meta);
  }

  var assignment = _astTypes.builders.assignmentExpression('=', variable, mapExpression(node.value, meta));

  if (node.context === '||=') {
    return _astTypes.builders.logicalExpression('||', variable, assignment);
  } else if (node.context === '?=') {
    return _astTypes.builders.conditionalExpression(_astTypes.builders.binaryExpression('!=', variable, _astTypes.builders.identifier('null')), variable, assignment);
  }

  assignment.operator = node.context || node.operator || '=';

  return assignment;
}

function mapObjectPatternItem(node, meta) {
  var type = node.constructor.name;
  if (type === 'Value') {
    return mapLiteral(node, meta);
  } else if (type === 'Assign') {
    if (node.value.base.properties) {
      return mapObjectPattern(node.value.base.properties, meta);
    }
    return mapExpression(node.value, meta);
  }

  throwError(node.locationData, 'can\'t convert node of type: ' + type + ' to ObjectPatternItem - not recognized');
}

function mapObjectPattern(nodes, meta) {
  return _astTypes.builders.objectPattern(nodes.map(function (node) {
    var operatorToken = node.operatorToken;

    var prop = undefined;
    prop = _astTypes.builders.property('init', mapKey(node.variable || node, meta), mapObjectPatternItem(node, meta));

    if (operatorToken === undefined) {
      prop.shorthand = true;
    }
    return prop;
  }));
}

function mapArrayPattern(node, meta) {
  return _astTypes.builders.arrayPattern(node.objects.map(function (prop) {
    var type = prop.base.constructor.name;
    if (type === 'Literal') {
      return mapLiteral(prop, meta);
    } else if (type === 'Arr') {
      return mapArrayPattern(prop.base, meta);
    } else if (type === 'Obj') {
      return mapObjectPattern(prop.base.properties, _extends({}, meta, { left: true }));
    }
  }));
}

function mapAssignmentPattern(node, meta) {
  // that's a destructuring assignment
  var type = node.constructor.name;

  if (type === 'Obj' && node.properties) {
    return mapObjectPattern(node.properties, meta);
  } else if (type === 'Arr') {
    return mapArrayPattern(node, meta);
  }

  return mapExpression(node, meta);
}

function coffeeParse(source) {
  var ast = (0, _coffeeScript.nodes)(source);
  return ast;
}

function transpile(ast, meta) {
  if (meta === undefined) {
    meta = {};
  }

  if (!meta.scope) {
    meta.scope = new _scope.Scope(null, ast, null, []);
    meta.indent = ' ';
  }

  var program = mapBlockStatement(ast, meta, _astTypes.builders.program);

  var utilities = meta.scope.utilities;

  var utils = Object.keys(utilities);

  utils.forEach(function (util) {
    var expr = _recast2.default.parse(utilities[util] + ' = ' + UTILITIES[util](meta)).program.body[0];

    delete expr.loc;

    program.body.unshift(expr);
  });

  return program;
}

// Not sure why this is happening just yet,
// backslashes gallore that is :(
function removeDoubleEscapes(compiled) {
  compiled.code = compiled.code.replace('\\\\\\', '\\');
  return compiled;
}

function compile(source, opts) {
  var parse = arguments.length <= 2 || arguments[2] === undefined ? coffeeParse : arguments[2];

  var doubleSemicolon = /\;+/g;
  opts = opts || { tabWidth: 2, quote: 'double' };

  var _compile = (0, _compose2.default)(
  // hack because of double semicolon
  removeDoubleEscapes, function (compiledSource) {
    return _extends({}, compiledSource, { code: compiledSource.code.replace(doubleSemicolon, ';') });
  }, function (jsAst) {
    return _recast2.default.print(jsAst, opts);
  }, insertSuperCalls, insertBreakStatements, insertVariableDeclarations, function (csAst) {
    return transpile(csAst, { options: opts });
  }, parse);

  return _compile(source).code;
}

var UTILITIES = {
  modulo: function modulo() {
    return 'function (a, b) { return (+a % (b = +b) + b) % b; }';
  },
  extend: function extend(o) {
    return 'function(child, parent) {\n  for (var key in parent) {\n    if (' + utility('hasProp', o) + '.call(parent, key)) child[key] = parent[key];\n  }\n  function ctor() {\n    this.constructor = child;\n  }\n  ctor.prototype = parent.prototype;\n  child.prototype = new ctor();\n  child.__super__ = parent.prototype;\n  return child;\n}';
  },
  hasProp: function hasProp() {
    return '{}.hasOwnProperty';
  }
};

// Helper for ensuring that utility functions are assigned at the top level.
// copied from coffee-script compiler
function utility(name, o) {
  var root = o.scope.root;


  if (root.utilities[name]) {
    return root.utilities[name];
  }

  var ref = root.freeVariable(name);
  root.assign(ref, UTILITIES[name](o));
  root.utilities[name] = ref;
}
